// Copyright (c) 2014, The Shelf Route project authors.
// Please see the AUTHORS file for details.
// All rights reserved. Use of this source code is governed by
// a BSD 2-Clause License that can be found in the LICENSE file.

library shelf_route;

import 'package:shelf/shelf.dart';
import 'package:path/path.dart' as p;
import 'package:matcher/matcher.dart';
import 'src/preconditions.dart';
import 'package:uri/uri.dart';
import 'package:shelf_path/shelf_path.dart';
export 'package:shelf_path/shelf_path.dart' show getPathParameters,
    getPathParameter;



/**
 * A [Handler] that routes to other handlers based on the incoming request
 * (path and method). Note there can be several layers of routers to facilitate
 * modularity.
 *
 * When the Router routes a request to a handler it adjusts the requests as follows:
 *
 * * the routes path is removed from the start of the requests pathInfo
 * * the routes path is added to the end of the requests scriptName
 *
 * e.g if original request had path info of `/banking` and scriptName was
 * `/abc` then when routing a request of `/banking/accounts` the new request
 * passed to the handler will have a pathInfo of `/accounts` and scriptName of
 * `/abc/banking`
 *
 * If [fallbackHandler] is provided it will be called if no routes match. If
 * omitted a default handler which returns a 404 response is used
 *
 * If [middleware] is included a pipeline will be created with the middleware
 * and [handler]
*/

Router router({ HandlerAdapter handlerAdapter: _noopHandlerAdapter,
  PathAdapter pathAdapter: uriTemplatePattern, Function fallbackHandler,
  Middleware middleware}) =>
    new Router(handlerAdapter: handlerAdapter, pathAdapter: pathAdapter,
      fallbackHandler: fallbackHandler, middleware: middleware);

/// Used to specify that the Router should route requests to all HTTP methods
/// for a given route
const ALL_METHODS = null;


typedef void Printer(String s);

/// prints out a representation of all the routes within the router, using
/// the [printer] if provided or the standard Dart [print] function if not.
///
void printRoutes(Router router, { Printer printer: print }) {
  router.fullPaths.forEach((pi) {
    printer('${pi.method}\t->\t${pi.path}');
  });
}

final _url = p.url;

class PathInfo {
  final String method;
  final String path;

  PathInfo(String method, String path)
      : this.method = method != null ? method : '*',
          this.path = path != null && !path.isEmpty ? path : '/';
}

typedef Handler HandlerAdapter(Function handler);

Handler _noopHandlerAdapter(Handler handler) =>
    handler;

/// A [HanderAdapter] that can be merged with another handlerAdapter of the same
/// type
abstract class MergeableHandlerAdpater {
  HandlerAdapter merge(MergeableHandlerAdpater other);
}

typedef UriPattern PathAdapter(path);

PathAdapter _wrapPathAdapter(PathAdapter pathAdapter) {
  return (path) {
    ensure(path, anyOf(isNull, new isInstanceOf<String>(),
        new isInstanceOf<UriPattern>()));

    if (path != null && path is UriPattern) {
      return path;
    }

    return pathAdapter(path);
  };

}

UriPattern uriTemplatePattern(path) {

  // Note: we change '/' to '' as UriParser strips trailing '/'
  // and a '' path won't match agains a pattern of '/'
  final nonNullPath = path == null ? '' : path == '/' ? '' : path;
  return new UriParser(new UriTemplate(nonNullPath));
}


class Router extends _BaseRoute{
  final List<_Route> _routes = <_Route>[];
  final Handler _fallbackHandler;

  final HandlerAdapter handlerAdapter;
  final PathAdapter pathAdapter;

  Router._internal(this._fallbackHandler, this.handlerAdapter,
      PathAdapter pathAdapter, path, Middleware middleware)
      : this.pathAdapter = pathAdapter,
        super(pathAdapter(path), middleware);

  // TODO: we could expose 'path' at this level if it makes sense but
  // would require further refactoring first as the Router's _handler
  // will ignore the path
  Router({Function fallbackHandler,
    HandlerAdapter handlerAdapter: _noopHandlerAdapter,
    PathAdapter pathAdapter: uriTemplatePattern,
    Middleware middleware})
      : this._internal(handlerAdapter(fallbackHandler != null ?
          fallbackHandler : _send404),
          handlerAdapter, _wrapPathAdapter(pathAdapter), '/', middleware);

  /**
   * Adds a route with the given [handler], [path] and [method].
   *
   * [path] may be either a String or a [UriPattern]. If path is a String it
   * will be parsed as a [UriParser] which means it is expected to conform to
   * a [UriTemplate].
   *
   * The [handler] must be of type [Handler] unless a custom [handlerAdapter]
   * has been set for this router. The [handlerAdapter] will be applied to
   * the provided [handler] to create the actual handler used.
   *
   * If [exactMatch] is true then a path such as `/foo/bar/fum` will not match
   * a path like `/foo/bar`, where as when it is false it will.
   *
   * If [middleware] is included a pipeline will be created with the middleware
   * and [handler]
   *
   * If [handlerAdapter] is provided it will be merged with the parent adapter
   * if it exists and both are of type [MergeableHandlerAdapter]. This allows
   * custom tweaking of creating the handler per route.
   */
  void add(dynamic path, List<String> methods, Function handler,
      { bool exactMatch: true, Middleware middleware,
        HandlerAdapter handlerAdapter }) {
    final ha = _maybeMergeHandlerAdapters(handlerAdapter);

    _routes.add(new _SimpleRoute(ha(handler), pathAdapter(path),
        methods, exactMatch, middleware));

  }

  /// adds all the routes created by [routeable]'s [createRoutes] method
  /// to the current router if [path] is not provided or to a new child [Router]
  /// at [path].
  ///
  /// If [middleware] is included a pipeline will be created with the middleware
  /// and [handler]
  void addAll(RouteableFunction routeable,
              { dynamic path, Middleware middleware,
                HandlerAdapter handlerAdapter }) {
    final r = path != null ?
        child(path, middleware: middleware,
            handlerAdapter: handlerAdapter) : this;
    routeable(r);
  }

  /// creates a route with method = GET
  void get(dynamic path, Function handler, { Middleware middleware,
      HandlerAdapter handlerAdapter }) {
    add(path, ['GET'], handler, middleware: middleware,
        handlerAdapter: handlerAdapter);
  }

  /// creates a route with method = POST
  void post(dynamic path, Function handler, { Middleware middleware,
      HandlerAdapter handlerAdapter }) {
    add(path, ['POST'], handler, middleware: middleware,
        handlerAdapter: handlerAdapter);
  }

  /// creates a route with method = PUT
  void put(dynamic path, Function handler, { Middleware middleware,
      HandlerAdapter handlerAdapter }) {
    add(path, ['PUT'], handler, middleware: middleware,
        handlerAdapter: handlerAdapter);
  }

  /// creates a route with method = DELETE
  void delete(dynamic path, Function handler, { Middleware middleware,
      HandlerAdapter handlerAdapter }) {
    add(path, ['DELETE'], handler, middleware: middleware,
        handlerAdapter: handlerAdapter);
  }

  /// adds a child [Router] at [path]
  Router child(dynamic path, { Middleware middleware,
      HandlerAdapter handlerAdapter, PathAdapter pathAdapter }) {
    final ha = _maybeMergeHandlerAdapters(handlerAdapter);

    final pa = pathAdapter != null ? _wrapPathAdapter(pathAdapter) :
      this.pathAdapter;

    final child = new Router._internal(_fallbackHandler, ha, pa, path, middleware);
    _routes.add(child);
    return child;
  }

  HandlerAdapter _maybeMergeHandlerAdapters(
      HandlerAdapter handlerAdapter) {
    var ha = null;
    if (handlerAdapter == null) {
      ha = this.handlerAdapter;
    }
    else if (handlerAdapter is MergeableHandlerAdpater &&
          this.handlerAdapter is MergeableHandlerAdpater) {
      MergeableHandlerAdpater haChild =
          handlerAdapter as MergeableHandlerAdpater;
      MergeableHandlerAdpater haParent =
          this.handlerAdapter as MergeableHandlerAdpater;

      ha = haChild.merge(haParent);
    }
    else {
      ha = handlerAdapter;
    }

    return ha;
  }

  /// adds a child [Router] at [path]
  /// WARNING: Experimental and untested. Use at your own peril ;-)
  void attach(dynamic path, Router childRouter, { Middleware middleware,
      HandlerAdapter handlerAdapter, PathAdapter pathAdapter }) {
    final c = child(path, middleware: middleware, handlerAdapter: handlerAdapter,
        pathAdapter: pathAdapter);
    c._routes.add(childRouter);
  }

  // TODO: this won't work if top level has a path
  Handler get _basicHandler => _handleChildRoutes;

  Iterable<PathInfo> get fullPaths => _createFullPaths([], null);

  Iterable<PathInfo> _createFullPaths(Iterable<String> rootPathSegs,
                                      String method) {
    final ourPaths = super._createFullPaths(rootPathSegs, method)
        .map((pi) => pi.path);

    return _routes.map((r) =>
        r._createFullPaths(ourPaths, method))
          .expand((e) => e);
  }

//  @override
//  bool _canHandle(Request request) =>
//      super._canHandle(request) && _routes.any((r) => r._canHandle(request));


  dynamic _handleChildRoutes(Request request) {
    final route = _routes.firstWhere((r) =>
        r._canHandle(request.method, request.url),
        orElse: () => null);
    if (route != null) {
      return route._handle(request) ;
    }
    else {
      return _fallbackHandler(request);
    }
  }


  @override
  List<String> get _methods => null;

  @override
  bool get _exactMatch => false;


  String toString() =>
      'Router at $_pathPattern; # routes ${_routes.length}';

}

/// A function capable of creating routes in a given [Router]
typedef RouteableFunction(Router router);

/// A class capable of creating routes in a given [Router].
/// In most cases `RouteableFunction` is simpler and more compact
abstract class Routeable {
  void createRoutes(Router router);
  void call(Router router) => createRoutes(router);
}

// base class for the _Route Composite Pattern
abstract class _Route {
  Handler get handler;

  Handler get _basicHandler;
  List<String> get _methods;
  bool get _exactMatch;

  bool _canHandle(String method, Uri url);
  dynamic _handle(Request request);

  Iterable<PathInfo> _createFullPaths(Iterable<String> rootPathSegs,
                                      String method);

}

abstract class _BaseRoute implements _Route {
  final UriPattern _pathPattern;
  Handler get _basicHandler;
  Handler _handler;
  Handler get handler => _handler;

  _BaseRoute(this._pathPattern, Middleware middleware) {
    ensure(_pathPattern, isNotNull, '_pathPattern cannot be null');
    _handler =  middleware != null ?
        const Pipeline().addMiddleware(middleware).addHandler(_basicHandler) :
          _basicHandler;
  }


  @override
  bool _canHandle(String method, Uri url) {
    if (_methods != null && !_methods.contains(method)) {
      return false;
    }

    final patternMatches = _pathPattern.matches(url);

    if (!patternMatches) {
      return false;
    }

    if (!_exactMatch) {
      return true;
    }

    final uriMatch = _pathPattern.match(url);
    final remainingSegments = uriMatch.rest.pathSegments;

    return remainingSegments.length == 0
      || (remainingSegments.length == 1
        && (remainingSegments.first == '' || remainingSegments.first == '/')
    );
  }

  @override
  dynamic _handle(Request request) {
    // create a new request with the pathInfo and scriptName adjusted relative
    // to route path
    final newRequest = _createNewRequest(request);

    return handler(newRequest);
  }

  Request _createNewRequest(Request request) {
    var url = request.url;

    final uriMatch = _pathPattern.match(url);

    final requestSegs = url.pathSegments;

    final skipCount = request.scriptName.isNotEmpty && requestSegs.isNotEmpty
        && requestSegs.first == '/' ? 1 : 0;
    final remainingPath = _cleanRest(uriMatch.rest);

    final extraScriptPath = requestSegs.skip(skipCount).take(
        requestSegs.length - remainingPath.pathSegments.length);
    var newScriptName = _adjustPath(_url.joinAll(_url.split(request.scriptName)
        ..addAll(extraScriptPath)));

    var newPathInfo = _adjustPath(remainingPath.toString());

    if (_exactMatch && newPathInfo == url.toString()) {
      newPathInfo = '';
    }

    if (newPathInfo == '' && newScriptName == '') {
      newPathInfo = '/';
    }

    final newUrl = Uri.parse(newPathInfo);

    var newContext = request.context;
    if (uriMatch.parameters.isNotEmpty) {
      newContext = {}..addAll(request.context);
      addPathParametersToContext(newContext, uriMatch.parameters);
    }

    return request.change(context: newContext, url: newUrl,
        scriptName: newScriptName);

  }

  // workaround for https://github.com/google/uri.dart/issues/16
  Uri _cleanRest(Uri rest) {
    if (rest.hasAuthority || rest.fragment == '' || rest.query == '') {
      final builder = new UriBuilder.fromUri(rest);
      if (builder.fragment == '') {
        builder.fragment = null;
      }
      if (builder.host == '') {
        builder.host = null;
      }
      if (builder.queryParameters != null && builder.queryParameters.isEmpty) {
        builder.queryParameters = null;
      }

      return builder.build();
    }
    return rest;
  }

  String _adjustPath(String path) {
    if (path == '/') {
      return '';
    }
    if (path.isNotEmpty && !path.startsWith('/') && !path.startsWith('?')) {
      return _url.join('/', path);
    }

    return path;
  }

  Iterable<PathInfo> _createFullPaths(Iterable<String> rootPathSegs,
                                      String method) {
    final routeSegsStr = (_pathPattern is UriParser)
      ? (_pathPattern as UriParser).template.template
      : "......";

    // remove '/' unless first segment
    final relRouteSegsStr = rootPathSegs.isEmpty ? routeSegsStr :
      _relative(routeSegsStr);


    final newPathSegs = new List()..addAll(rootPathSegs)..add(relRouteSegsStr);

    final path = (_url.joinAll(newPathSegs) as String)
        .replaceAll('/{?', '{?'); // ugly

    final _newMethods = method != null ? [method] :
      _methods != null ? _methods : ['*'];
    return _newMethods.map((m) => new PathInfo(m, path));

  }

  String _relative(String routeSegsStr) {
    final sep = _url.separator;
    return routeSegsStr.startsWith(sep) ? routeSegsStr.substring(1)
        : routeSegsStr;
  }


}

class _SimpleRoute extends _BaseRoute {
  final Handler _basicHandler;
  final List<String> _methods;
  final bool _exactMatch;

  _SimpleRoute(this._basicHandler, UriPattern _pathPattern, this._methods,
      this._exactMatch, Middleware middleware)
      : super(_pathPattern, middleware) {
    ensure(_basicHandler, isNotNull, 'handler cannot be null');
  }


  String toString() =>
      '$_methods->$_pathPattern ${_exactMatch ? '(exact)' : ''}';

}

Response _send404(Request req) {
  return new Response.notFound("Not Found");
}


